@@ -17,7 +17,7 @@
 from concurrent.futures import ProcessPoolExecutor, CancelledError
 from enum import Enum
 from io import StringIO
-from mesonbuild._pathlib import Path, PurePath
+from pathlib import Path, PurePath
 import argparse
 import functools
 import itertools
@@ -42,7 +42,8 @@ from mesonbuild import mesonlib
 from mesonbuild import mlog
 from mesonbuild import mtest
 from mesonbuild.build import ConfigurationData
-from mesonbuild.mesonlib import MachineChoice, Popen_safe
+from mesonbuild.mesonlib import MachineChoice, Popen_safe, TemporaryDirectoryWinProof
+from mesonbuild.mlog import bold, green, red, yellow
 from mesonbuild.coredata import backendlist, version as meson_version
 
 from run_tests import get_fake_options, run_configure, get_meson_script
@@ -224,25 +225,9 @@ class TestDef:
             return (s_id, self.path, self.name or '') < (o_id, other.path, other.name or '')
         return NotImplemented
 
-class AutoDeletedDir:
-    def __init__(self, d):
-        self.dir = d
-
-    def __enter__(self):
-        os.makedirs(self.dir, exist_ok=True)
-        return self.dir
-
-    def __exit__(self, _type, value, traceback):
-        # We don't use tempfile.TemporaryDirectory, but wrap the
-        # deletion in the AutoDeletedDir class because
-        # it fails on Windows due antivirus programs
-        # holding files open.
-        mesonlib.windows_proof_rmtree(self.dir)
-
 failing_logs = []
 print_debug = 'MESON_PRINT_TEST_OUTPUT' in os.environ
 under_ci = 'CI' in os.environ
-under_xenial_ci = under_ci and ('XENIAL' in os.environ)
 skip_scientific = under_ci and ('SKIP_SCIENTIFIC' in os.environ)
 do_debug = under_ci or print_debug
 no_meson_log_msg = 'No meson-log.txt found.'
@@ -361,22 +346,6 @@ def log_text_file(logfile, testdir, stdo, stde):
         raise StopException()
 
 
-def bold(text):
-    return mlog.bold(text).get_text(mlog.colorize_console())
-
-
-def green(text):
-    return mlog.green(text).get_text(mlog.colorize_console())
-
-
-def red(text):
-    return mlog.red(text).get_text(mlog.colorize_console())
-
-
-def yellow(text):
-    return mlog.yellow(text).get_text(mlog.colorize_console())
-
-
 def _run_ci_include(args: T.List[str]) -> str:
     if not args:
         return 'At least one parameter required'
@@ -520,14 +489,17 @@ def detect_parameter_files(test: TestDef, test_build_dir: str) -> (Path, Path):
 def run_test(test: TestDef, extra_args, compiler, backend, flags, commands, should_fail, use_tmp: bool):
     if test.skip:
         return None
-    with AutoDeletedDir(create_deterministic_builddir(test, use_tmp)) as build_dir:
-        with AutoDeletedDir(tempfile.mkdtemp(prefix='i ', dir=None if use_tmp else os.getcwd())) as install_dir:
+    build_dir = create_deterministic_builddir(test, use_tmp)
+    try:
+        with TemporaryDirectoryWinProof(prefix='i ', dir=None if use_tmp else os.getcwd()) as install_dir:
             try:
                 return _run_test(test, build_dir, install_dir, extra_args, compiler, backend, flags, commands, should_fail)
             except TestResult as r:
                 return r
             finally:
                 mlog.shutdown() # Close the log file because otherwise Windows wets itself.
+    finally:
+        mesonlib.windows_proof_rmtree(build_dir)
 
 def _run_test(test: TestDef, test_build_dir: str, install_dir: str, extra_args, compiler, backend, flags, commands, should_fail):
     compile_commands, clean_commands, install_commands, uninstall_commands = commands
@@ -620,7 +592,7 @@ def _run_test(test: TestDef, test_build_dir: str, install_dir: str, extra_args,
 
     # Do installation, if the backend supports it
     if install_commands:
-        env = os.environ.copy()
+        env = test.env.copy()
         env['DESTDIR'] = install_dir
         # Install with subprocess
         pi, o, e = Popen_safe(install_commands, cwd=test_build_dir, env=env)
@@ -630,7 +602,7 @@ def _run_test(test: TestDef, test_build_dir: str, install_dir: str, extra_args,
             return testresult
 
     # Clean with subprocess
-    env = os.environ.copy()
+    env = test.env.copy()
     pi, o, e = Popen_safe(clean_commands + dir_args, cwd=test_build_dir, env=env)
     testresult.add_step(BuildStep.clean, o, e)
     if pi.returncode != 0:
@@ -666,6 +638,7 @@ def gather_tests(testdir: Path, stdout_mandatory: bool) -> T.List[TestDef]:
             env = test_def['env']
             for key, val in env.items():
                 val = val.replace('@ROOT@', t.path.resolve().as_posix())
+                val = val.replace('@PATH@', t.env.get('PATH', ''))
                 env[key] = val
 
         # Handle installed files
@@ -791,7 +764,7 @@ def have_d_compiler():
     return False
 
 def have_objc_compiler(use_tmp: bool) -> bool:
-    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir=None if use_tmp else '.')) as build_dir:
+    with TemporaryDirectoryWinProof(prefix='b ', dir=None if use_tmp else '.') as build_dir:
         env = environment.Environment(None, build_dir, get_fake_options('/'))
         try:
             objc_comp = env.detect_objc_compiler(MachineChoice.HOST)
@@ -807,7 +780,7 @@ def have_objc_compiler(use_tmp: bool) -> bool:
     return True
 
 def have_objcpp_compiler(use_tmp: bool) -> bool:
-    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir=None if use_tmp else '.')) as build_dir:
+    with TemporaryDirectoryWinProof(prefix='b ', dir=None if use_tmp else '.') as build_dir:
         env = environment.Environment(None, build_dir, get_fake_options('/'))
         try:
             objcpp_comp = env.detect_objcpp_compiler(MachineChoice.HOST)
@@ -828,8 +801,8 @@ def have_java():
     return False
 
 def skippable(suite, test):
-    # Everything is optional when not running on CI, or on Ubuntu 16.04 CI
-    if not under_ci or under_xenial_ci:
+    # Everything is optional when not running on CI
+    if not under_ci:
         return True
 
     if not suite.endswith('frameworks'):
@@ -871,6 +844,10 @@ def skippable(suite, test):
     if test.endswith('15 llvm'):
         return True
 
+    # This test breaks with gobject-introspection <= 1.58.1
+    if test.endswith('34 gir static lib'):
+        return True
+
     # No frameworks test should be skipped on linux CI, as we expect all
     # prerequisites to be installed
     if mesonlib.is_linux():
@@ -887,6 +864,10 @@ def skippable(suite, test):
     if test.endswith('4 qt') and mesonlib.is_osx():
         return False
 
+    # Bindgen isn't available in all distros
+    if test.endswith('12 bindgen'):
+        return False
+
     # Other framework tests are allowed to be skipped on other platforms
     return True
 
@@ -1187,7 +1168,7 @@ def check_meson_commands_work(options):
     global backend, compile_commands, test_commands, install_commands
     testdir = PurePath('test cases', 'common', '1 trivial').as_posix()
     meson_commands = mesonlib.python_command + [get_meson_script()]
-    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir=None if options.use_tmpdir else '.')) as build_dir:
+    with TemporaryDirectoryWinProof(prefix='b ', dir=None if options.use_tmpdir else '.') as build_dir:
         print('Checking that configuring works...')
         gen_cmd = meson_commands + [testdir, build_dir] + backend_flags + options.extra_args
         pc, o, e = Popen_safe(gen_cmd)
@@ -1217,7 +1198,7 @@ def check_meson_commands_work(options):
 def detect_system_compiler(options):
     global host_c_compiler, compiler_id_map
 
-    with AutoDeletedDir(tempfile.mkdtemp(prefix='b ', dir=None if options.use_tmpdir else '.')) as build_dir:
+    with TemporaryDirectoryWinProof(prefix='b ', dir=None if options.use_tmpdir else '.') as build_dir:
         fake_opts = get_fake_options('/')
         if options.cross_file:
             fake_opts.cross_file = [options.cross_file]
